""" Module defining constants """

# string constants for processor names
LANGID = 'langid'
TOKENIZE = 'tokenize'
MWT = 'mwt'
POS = 'pos'
LEMMA = 'lemma'
DEPPARSE = 'depparse'
NER = 'ner'
SENTIMENT = 'sentiment'
CONSTITUENCY = 'constituency'
